const   BUF_END   = $3FFF;
        MAX_PARAM = 10;
        REC_LEN = $80;
        EMPTY = $FF;

type    NUM_STRING  = string [6];
        NAME_STRING = string [16];
        TNR_STRING  = string [4];
        VPP_STRING  = string [6];
        LINE        = string [40];
        EPROM_DATA  = record
          TYP   : NAME_STRING;
          LEN   : integer;
          TIME  : integer;
          TNR   : TNR_STRING;
          VPP   : VPP_STRING;
          PGMD  : char;
          COMMENT: string [7];
          RD, VER, PROG, PINH : array [3..4] of byte;
        end;

var     ERROR, I, J, K : real;
        FILE1 : file;
        FILE1_NAME : NAME_STRING;
        CMD_LINE : LINE;
        PARAM : array [1..MAX_PARAM] of NAME_STRING;
        PARAM_NO : integer;
        H1, H2, H3 ,H4, H5, H6 , N1: real;
        E1, E2, E3,  N2, N3 ,N4 , W: integer;
        error_no: integer;
        BLOCK_BUF : array [0..REC_LEN] of byte;
        EPROM : EPROM_DATA;
        EPROMH : EPROM_DATA;
        DATA_EPROM, DATA_RAM : byte;
        CONTINUE : boolean;
        EPROM_ERROR_FLAG : boolean;
        LAUF : boolean;
        BUFFER : array [0..BUF_END] of byte;



const   EPROM_TYPES = 15;
        KNOWN_EPROMS : array [1..EPROM_TYPES] of EPROM_DATA =

         ((TYP  : '2716     '       ;     LEN  : $0800    ;    TIME :  1          ;
           TNR  : '10'              ;     VPP  : '25 V  ' ;    PGMD : 'C'         ;
           COMMENT:'ALL   '         ;     RD   : ($28,$00);    VER  : ($20,$78)   ;
                                          PROG : ($20,$7F);    PINH : ($20,$1E)  ),

          (TYP  : '2732     '       ;     LEN  : $1000    ;    TIME :  1          ;
           TNR  : '20'              ;     VPP  : '25 V  ' ;    PGMD : 'C'         ;
           COMMENT:'ALL   '         ;     RD   : ($20,$00);    VER  : ($20,$18)   ;
                                          PROG : ($20,$3C);    PINH : ($20,$39)  ),

          (TYP  : '2732A    '       ;     LEN  : $1000    ;    TIME :  1          ;
           TNR  : '30'              ;     VPP  : '21 V  ' ;    PGMD : 'C'         ;
           COMMENT:'ALL   '         ;     RD   : ($20,$00);    VER  : ($20,$00)   ;
                                          PROG : ($20,$24);    PINH : ($20,$21)  ),

          (TYP  : '2764     '       ;     LEN  : $2000    ;    TIME :  1          ;
           TNR  : '40'              ;     VPP  : '21 V  ' ;    PGMD : 'C'         ;
           COMMENT:'ALL   '         ;     RD   : ($C0,$10);    VER  : ($40,$50)   ;
                                          PROG : ($00,$56);    PINH : ($40,$13)  ),

          (TYP  : '2764     '       ;     LEN  : $2000    ;    TIME : 1           ;
           TNR  : '50'              ;     VPP  : '12,5 V' ;    PGMD : 'C'         ;
           COMMENT:'ALL   '         ;     RD   : ($C0,$10);    VER  : ($40,$C8)   ;
                                          PROG : ($00,$CE);    PINH : ($40,$8B)  ),


          (TYP  : '27128    '       ;     LEN  : $4000    ;    TIME : 1           ;
           TNR  : '60'              ;     VPP  : '12,5 V' ;    PGMD : 'C'         ;
           COMMENT:'ALL   '         ;     RD   : ($C0,$10);    VER  : ($40,$C8)   ;
                                          PROG : ($00,$CE);    PINH : ($40,$8B)  ),


          (TYP  : '27128    '       ;     LEN  : $4000    ;    TIME : 1           ;
           TNR  : '65'              ;     VPP  : '21 V  ' ;    PGMD : 'C'         ;
           COMMENT:'ALL   '         ;     RD   : ($C0,$10);    VER  : ($40,$48)   ;
                                          PROG : ($00,$4E);    PINH : ($40,$0B)  ),


          (TYP  : '27256    '       ;     LEN  : $4000    ;    TIME :  1          ;
           TNR  : '70'              ;     VPP  : '21 V  ' ;    PGMD : 'C'         ;
           COMMENT:'1.Teil'         ;     RD   : ($80,$10);    VER  : ($00,$49)   ;
                                          PROG : ($00,$56);    PINH : ($00,$0F)  ),

          (TYP  : '27256    '       ;     LEN  : $4000    ;    TIME : 1           ;
           TNR  : '71'              ;     VPP  : '21 V  ' ;    PGMD : 'C'         ;
           COMMENT:'2.Teil'         ;     RD   : ($C0,$10);    VER  : ($40,$49)   ;
                                          PROG : ($40,$56);    PINH : ($40,$0F)  ),


          (TYP  : '27256    '       ;     LEN  : $4000    ;    TIME : 1           ;
           TNR  : '75'              ;     VPP  : '12,5 V' ;    PGMD : 'C'         ;
           COMMENT:'1.Teil'         ;     RD   : ($80,$10);    VER  : ($00,$C9)   ;
                                          PROG : ($00,$CE);    PINH : ($00,$8F)  ),

          (TYP  : '27256    '       ;     LEN  : $4000    ;    TIME : 1           ;
           TNR  : '76'              ;     VPP  : '12,5 V' ;    PGMD : 'C'         ;
           COMMENT:'2.Teil'         ;     RD   : ($C0,$10);    VER  : ($40,$C9)   ;
                                          PROG : ($40,$CE);    PINH : ($40,$8F)  ),


          (TYP  : '27C256   '       ;     LEN  : $4000    ;    TIME : 1           ;
           TNR  : '80'              ;     VPP  : '12,5 V' ;    PGMD : 'D'         ;
           COMMENT:'1.Teil'         ;     RD   : ($80,$10);    VER  : ($00,$CB)   ;
                                          PROG : ($00,$CE);    PINH : ($00,$8F)  ),

          (TYP  : '27C256   '       ;     LEN  : $4000    ;    TIME : 1           ;
           TNR  : '81'              ;     VPP  : '12,5 V' ;    PGMD : 'D'         ;
           COMMENT:'2.Teil'         ;     RD   : ($C0,$10);    VER  : ($40,$CB)   ;
                                          PROG : ($40,$CE);    PINH : ($40,$8F)  ),

          (TYP  : '27C256   '       ;     LEN  : $4000    ;    TIME : 1           ;
           TNR  : '85'              ;     VPP  : '21 V  ' ;    PGMD : 'D'         ;
           COMMENT:'1.Teil'         ;     RD   : ($80,$10);    VER  : ($00,$4B)   ;
                                          PROG : ($00,$4E);    PINH : ($00,$0F) ) ,

          (TYP  : '27C256   '       ;     LEN  : $4000    ;    TIME : 1           ;
           TNR  : '86'              ;     VPP  : '21 V  ' ;    PGMD : 'D'         ;
           COMMENT:'2.Teil'         ;     RD   : ($C0,$10);    VER  : ($40,$4B)   ;
                                          PROG : ($40,$4E);    PINH : ($40,$0F) ));


        PORT_OUT = $1d;
        PORT_IN =  $1d;
        PORT_STR = $02;
        PORT_CLK = $04;
        PORT_OUT_DAT = $08;
        PORT_IN_DAT = $01;
        PORT_CONTROLL= $1f;
        PORT_DAT : byte = $00;
        NOT_PORT_CLK = $FB;

type    DATA_BYTES = array [1..4] of byte;

var     OUT_BYTE : DATA_BYTES;
        IN_BYTE : byte;

procedure INIT_PIO;
begin
  port [PORT_CONTROLL] := $ff;  {Bitbetrieb}
  port [PORT_CONTROLL] := $01;  {D0 = Eingabebit}
end;

procedure STROBE;
begin
  port [PORT_OUT] := PORT_DAT or PORT_STR;
  port [PORT_OUT] := PORT_DAT;
end;

procedure CLOCK;
begin
  port [PORT_OUT] := PORT_DAT or PORT_CLK;
  port [PORT_OUT] := PORT_DAT;
end;

procedure PUT_DATA;
begin
  inline (
  $21/OUT_BYTE/    {    LD   HL,OUT_BYTE      }
  $06/$04/         {    LD    B,4             }
  $56/             {L0: LD    D,(HL)          }
  $0E/$08/         {    LD    C,8             }
  $CB/$3A/         {L1: SRL   D               }
  $3E/PORT_CLK/    {    LD    A,PORT_CLK      }
  $30/02/          {    JR   NC,L2            }
  $F6/PORT_OUT_DAT/{    OR    PORT_OUT_DAT    }
  $D3/PORT_OUT/    {S1: OUT   PORT_OUT,A      }
  $E6/NOT_PORT_CLK/{    AND    NOT PORT_CLK   }
  $D3/PORT_OUT/    {    OUT    PORT_OUT,A     }
  $0D/             {    DEC    C              }
  $20/$EF/         {    JR     NZ,L1          }
  $23/             {    INC    HL             }
  $05/             {    DEC    B              }
  $20/$E8/         {    JR     NZ,L0          }
  $3E/PORT_STR/    {    LD     A,PORT_STR     }
  $D3/PORT_OUT/    {    OUT    PORT_OUT       }
  $AF/             {    XOR    A              }
  $D3/PORT_OUT);   {    OUT    PORT_OUT       }
end;



function GET_DATA : byte;
var     IN_BYTE : byte;
begin
  inline (
  $06/$00/         {    LD  B,0               }
  $0E/$08/         {    LD  C,8               }
  $DB/PORT_IN/     {L1: IN  A,PORT_IN         }
  $E6/PORT_IN_DAT/ {    AND PORT_IN_DAT       }
  $28/$01/         {    JR  Z,S1              }
  $37/             {    SCF                   }
  $CB/$18/         {S1: RR B                  }
  $3E/PORT_CLK/    {    LD A,PORT_CLK         }
  $D3/PORT_OUT/    {    OUT PORT_OUT,A        }
  $E6/NOT_PORT_CLK/{    AND NOT_PORT_CLK      }
  $D3/PORT_OUT/    {    OUT PORT_OUT,A        }
  $0D/             {    DEC C                 }
  $20/$EC/         {    JR NZ,L1              }
  $21/IN_BYTE/     {    LD  HL,IN_BYTE        }
  $70);            {    LD  (HL),B            }
  GET_DATA := IN_BYTE;
end;


procedure RESET_SOCKET;
const   INIT_3 : DATA_bytes = ($00, $00, $00, $00);
begin
  OUT_BYTE := INIT_3;
  PUT_DATA;
end;

procedure INIT_HARDWARE;
const   INIT_1 : DATA_BYTES = ($55, $00, $00, $04);
        INIT_2 : DATA_BYTES = ($AA, $00, $00, $04);
var     INIT_OK : boolean;
        C : char;
begin
STROBE; CLOCK;
  INIT_OK := false;
  while keypressed do read (kbd, C);
  repeat
    OUT_BYTE := INIT_1; PUT_DATA;
    if GET_DATA = OUT_BYTE [1] then begin
      OUT_BYTE := INIT_2; PUT_DATA;
      if GET_DATA = OUT_BYTE [1] then begin
        RESET_SOCKET;
        INIT_OK := true;
      end;
    end;
    if not INIT_OK then begin
      gotoxy(1,15);
      clreol;
      write ('                         * bitte multiprom anschliessen *');
      delay(500);
      if keypressed then begin
        read (kbd, C);
        if C = #13 then begin
          INIT_OK := true;
          clrscr;
          gotoxy(1,13);
          write ('                 * kein Zugriff auf multiprom Hardware moeglich *');
        end;
      end
      else begin
      gotoxy(1,15);
      clreol;
      write ('                           * <Return> Keine Hardware *');
      delay(500);
      end;
      if keypressed then begin
        read (kbd, C);
        if C = #13 then begin
          INIT_OK := true;
          clrscr;
          gotoxy(1,13);
          write (chr ($07),'                 * kein Zugriff auf multiprom Hardware moeglich *');
        end;
      end;
    end;
  until INIT_OK;
  if C <> #13 then begin
    clrscr;
    gotoxy(1,13);
    write ('                           * multiprom hardware ok *');
  end;
  delay (1000);
end;

procedure PROG_VOLTAGE_ON;
begin
  OUT_BYTE [3] := EPROM.PINH [3];
  OUT_BYTE [4] := EPROM.PINH [4];
  PUT_DATA;
  delay (500);
end;

function GET_EPROM (ADR1 : real) : byte;
var ADR : integer;
begin;
  ADR := trunc (ADR1);
  OUT_BYTE [2] := lo (ADR);
  OUT_BYTE [3] := hi (ADR) or EPROM.RD [3];
  OUT_BYTE [4] := EPROM.RD [4];
  PUT_DATA;
  GET_EPROM := GET_DATA;
end;

function PUT_EPROM (ADR1 : real; DATA : byte) : byte;
var ADR : integer;
begin
  ADR := trunc (ADR1);
  OUT_BYTE [2] := lo (ADR);
  if DATA <> $FF then
  begin
    OUT_BYTE [1] := DATA;
    OUT_BYTE [3] := hi (ADR) or EPROM.PROG [3];
    OUT_BYTE [4] := EPROM.PROG [4];
    PUT_DATA;
       
    delay (EPROM.TIME);
    
    {H5 := 1; H5 := 0;  Zur Verzoegerung, ergibt fuer EPROM.TIME = 0 ,tpw = 1ms}
  end;
  OUT_BYTE [3] := hi (ADR) or EPROM.VER [3];
  OUT_BYTE [4] := EPROM.VER [4];
  PUT_DATA;
  PUT_EPROM := GET_DATA;
end;

function PUT_DEPROM (ADR1 : real; DATA : byte) : byte;
var ADR : integer;
begin
  ADR := trunc (ADR1);
  OUT_BYTE [2] := lo (ADR);
  if DATA <> $FF then
  begin
    OUT_BYTE [1] := DATA;
    OUT_BYTE [3] := hi (ADR) or EPROM.PROG [3];
    OUT_BYTE [4] := EPROM.PROG [4];
    PUT_DATA;
       
    delay (EPROM.TIME);
    
    H5 := 1; H5 := 0;  {Zur Verzoegerung, ergibt fuer EPROM.TIME = 0 ,tpw = 1ms}
  end;
  OUT_BYTE [3] := hi (ADR) or EPROM.VER [3];
  OUT_BYTE [4] := EPROM.VER [4];
  PUT_DATA;
  OUT_BYTE [4] := EPROM.VER [4]-3;
  PUT_DATA;
  PUT_DEPROM := GET_DATA;
  OUT_BYTE [4] := EPROM.VER [4];
  PUT_DATA;
end;

function PUT_OEVPPEPROM (ADR1 : real; DATA : byte) : byte;
var ADR : integer;
begin
  ADR := trunc (ADR1);
  OUT_BYTE [2] := lo (ADR);
  if DATA <> $FF then
  begin
    OUT_BYTE [1] := DATA;
    OUT_BYTE [3] := hi (ADR) or EPROM.PROG [3];
    OUT_BYTE [4] := EPROM.PROG [4];
    PUT_DATA;
       
    delay (EPROM.TIME);
    
    H5 := 1; H5 := 0;  {Zur Verzoegerung, ergibt fuer EPROM.TIME = 0 ,tpw = 1ms}
  end;
  OUT_BYTE [3] := hi (ADR) or EPROM.VER [3];
  OUT_BYTE [4] := EPROM.VER [4] + 1;
  PUT_DATA;
  OUT_BYTE [4] := EPROM.VER [4];
  PUT_DATA;
  PUT_OEVPPEPROM := GET_DATA;
end;

procedure PUTO_EPROM (ADR1 : real; DATA : byte) ;
var ADR : integer;
begin
  ADR := trunc (ADR1);
  OUT_BYTE [2] := lo (ADR);
  if DATA <> $FF then
  begin
    OUT_BYTE [1] := DATA;
    OUT_BYTE [3] := hi (ADR) or EPROM.PROG [3];
    OUT_BYTE [4] := EPROM.PROG [4];
    PUT_DATA;
       
    delay (EPROM.TIME);
    
    H5 := 1; H5 := 0;  {Zur Verzoegerung, ergibt fuer EPROM.TIME = 0 ,tpw = 1ms}
  end;
  OUT_BYTE [3] := hi (ADR) or EPROM.VER [3];
  OUT_BYTE [4] := EPROM.VER [4];
  PUT_DATA;
end;

function HEX (I : byte) : NUM_STRING;
const   HEX_NUMBERS : string[16] = '0123456789ABCDEF';
begin
  HEX := HEX_NUMBERS [(I shr 4) + 1]
       + HEX_NUMBERS [(I and $0F) + 1];
end;

function HEX_ADR (I : integer) : NUM_STRING;
begin
  HEX_ADR := HEX (hi (I)) + HEX (lo (I));
end;

function DECIMAL (H : NUM_STRING; var ERROR : integer) : real;
var     I : integer;
        D, VAL : real;
begin
  ERROR := 0;
  D := 0;
  for I := 1 to length (H) do
  begin
    case H [I] of
      '0'..'9' : VAL := ord (H [I]) - ord ('0');
      'A'..'F' : VAL := ord (H [I]) - ord ('A') + 10;
      else       VAL := 0;
                 ERROR := 1;
    end;
    D := D * 16 + VAL;
  end;
  if ERROR > 0
  then DECIMAL := 0
  else DECIMAL := D;
end; { DECIMAL }

function DECIMAL_CHECK (H : NUM_STRING; var ERROR : integer) : real;
var     D : real;
begin
  D := DECIMAL (H, ERROR);
  if D > BUF_END then
  begin
    DECIMAL_CHECK := 0;
    ERROR := 1;
  end
  else DECIMAL_CHECK := D;
end;
